local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("questsTooltip", "AceEvent-3.0")
local LQTo = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-QuestsTooltip")
local LibQTip = LibStub("LibQTip-1.0")
local dataobj = LibStub("LibDataBroker-1.1"):NewDataObject("tomQuest2", {type = "data source", icon = "Interface\\AddOns\\tomQuest2\\media\\icon"})


--------------------------------------
--- Copy global functions to local ---
--------------------------------------
local pairs = pairs
local format = string.format
local ToggleFrame = ToggleFrame
local ShowUIPanel = ShowUIPanel
local SetAbandonQuest = SetAbandonQuest
local SelectQuestLogEntry = SelectQuestLogEntry
local QuestLogPushQuest = QuestLogPushQuest
local MouseIsOver = MouseIsOver
local ItemRefTooltip = ItemRefTooltip
local GetQuestLogSelection = GetQuestLogSelection
local GetQuestLogPushable = GetQuestLogPushable
local GetQuestLink = GetQuestLink
local CreateFrame = CreateFrame
local CreateFont = CreateFont


local defaults = {
	profile = {
		showMaxQuest = true,
		showNumQuests = true,
		showNumCompletedQuests = true,
		hideCompletedObjective = 3,
		headerFontSize = 12,
		normalFontSize = 10,
		scale = 1,
		maxWidth = 300,
		showCurrentQuest = true,
		showCompletedQuest = true,
		showMaxQuest = true,
		questscellpadding = 5,
		objectivescellpadding = 10,
		chatlink = true,
		questsobjectives = true,
		clicktooltip = true,
		sharequest = true,
		trackquest = true,
		abandonquest = true,
		questlog = true,
		overtooltip = true,
		collapseButtonSize = 24,
		colorizeHeaderName = true,
		colorizeHeaderLevel = true,
		colorizeQuestName = true,
		colorizeQuestLevel = true,
		colorizeQuestObjective = true,
	}
}

local prdb, chdb
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("questsTooltip", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	chdb.collapseAll = chdb.collapseAll or 0
	chdb.zoneDisplay = chdb.zoneDisplay or {}
end

module.modName = LQTo["Broker Quests Tooltip"]

local options
function module:getOptions()
	if not options then
		options = {
			type = "group",
			name = module.modName,
			desc = module.modName,
			get = function(info) 
				return prdb[info[#info]]
			end,
			set = function(info, value)
				prdb[info[#info]] = value
				if info[#info] == "showNumCompletedQuests" or info[#info] == "showNumQuests" or info[#info] == "showMaxQuest" then
					module:updateLDBText()
				end
			end,
			args = {
				showNumCompletedQuests = {
					name = LQTo["Show number of completed quest"],
					type = "toggle",
					width = "full",
					order = 11,
				},
				showNumQuests = {
					name = LQTo["Show number of quests"],
					type = "toggle",
					width = "full",
					order = 12,
				},
				showMaxQuest = {
					name = LQTo["Show maximum quests number"],
					type = "toggle",
					width = "full",
					order = 13,
				},
				tooltipheader = {
					type = "header",
					name = LQTo["Tooltip Settings"],
					order = 14,
				},
				hideCompletedObjective = {
					name = LQTo["Hide completed quests Objectives"],
					type = "select",
					values = {[1] = LQTo["When you've completed the objective"], [2] = LQTo["When everyone in the party have completed the objective"], [3] = LQTo["Never"]},
					width = "full",
					order = 15,
				},
				headerFontSize = {
					name = LQTo["Quests font size"],
					type="range",
					min = 6,
					max = 20,
					step = 1,
					width = "full",
					order = 16,
				},
				normalFontSize = { 
					name = LQTo["Quests Objectives font size"],
					type="range",
					min = 6,
					max = 20,
					step = 1,
					width = "full",
					order = 17,
				},
				maxWidth = {
					name = LQTo["Quests Tooltip column max width"],
					type="range",
					min = 50,
					max = 600,
					step = 5,
					width = "full",
					order = 18,
				},
				scale = {
					name = LQTo["Quests Tooltip scale"],
					type="range",
					min = 0,
					max = 1,
					step = 0.1,
					width = "full",
					order = 19,
				},
				questscellpadding = {
					name = LQTo["Quests cell padding"],
					type="range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 20,
				},
				objectivescellpadding = {
					name = LQTo["Quests objectives cell padding"],
					type="range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 21,
				},
				collapseButtonSize = {
					name = LQTo["Collapse Button Size"],
					type = "range",
					min = 10,
					max = 40,
					step = 1,
					width = "full",
					order = 22,
				},
				colorizeHeaderName = {
					name = LQTo["Colorize Quests Header Name"],
					type = "toggle",
					width = "full",
					order = 23,
				},
				colorizeHeaderLevel = {
					name = LQTo["Colorize Quests Header Level"],
					type = "toggle",
					width = "full",
					order = 24,
				},
				colorizeQuestName = {
					name = LQTo["Colorize Quests Name"],
					type = "toggle",
					width = "full",
					order = 25,
				},
				colorizeQuestLevel = {
					name = LQTo["Colorize Quests Level"],
					type = "toggle",
					width = "full",
					order = 26,
				},
				colorizeQuestObjective = {
					name = LQTo["Colorize Quests Objectives"],
					type = "toggle",
					width = "full",
					order = 27,
				},
				tooltipinteraction = {
					type = "header",
					name = LQTo["Tooltip interaction"],
					order = 28,
				},
				chatlink  = {
					name = LQTo["Toggle chatlink when shift - right clicking on a quest"],
					type = "toggle",
					width = "full",
					order = 29,
				},
				questsobjectives = {
					name = LQTo["Toggle insert of quest objectives in chat when ctrl - right clicking on a quest"],
					type = "toggle",
					width = "full",
					order = 30,
				},
				clicktooltip = {
					name = LQTo["Toggle display of quest tooltip when right clicking on a quest"],
					type = "toggle",
					width = "full",
					order = 31,
				},
				sharequest = {
					name = LQTo["Toggle quest sharing when ctrl - left clicking on a quest"],
					type = "toggle",
					width = "full",
					order = 32,
				},
				trackquest = {
					name = LQTo["Toggle quest tracking when shift - left clicking on a quest"],
					type = "toggle",
					width = "full",
					order = 33,
				},
				abandonquest = {
					name = LQTo["Toggle quest abandon when alt - left clicking on a quest"],
					type = "toggle",
					width = "full",
					order = 34,
				},
				questlog = {
					name = LQTo["Toggle questlog display when left clicking on a quest"],
					type = "toggle",
					width = "full",
					order = 35,
				},
				overtooltip = {
					name = LQTo["Toggle quest tooltip when mouse over a quest"],
					type = "toggle",
					width = "full",
					order = 36,
				},
			},
		}
	end
	return options
end

local tip, tipAnchor
local tooltipIsShown = false
local maxQuests = MAX_QUESTS
local function hideTooltip()
	if not tip or MouseIsOver(tip) then return end
	tooltipIsShown = false
	tip:SetScript("OnLeave", nil)
	LibQTip:Release(tip)
	tip = nil
end

local timer = CreateFrame("Frame")

function module:OnEnable()
	core.RegisterCallback(self, "updateQuestsTooltip")
	core.RegisterCallback(self, "updateLDBText")
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
end

function module:updateQuestsTooltip(event)
	if tooltipIsShown then
		tip:SetScript("OnLeave", nil)
		LibQTip:Release(tip)
		tip = nil
		module:updateTooltip()
	end
end

function module:updateLDBText(event)
	core.numQuests = core.numQuests or 0
	core.numCompletedQuests = core.numCompletedQuests or 0
	if prdb.showNumCompletedQuests then
		if prdb.showNumQuests then
			if prdb.showMaxQuest then
				if core.numQuests > 0 then
					dataobj.text = format("|cff%s%s/%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedQuests/core.numQuests), core.numCompletedQuests, core.numQuests, maxQuests)
				else
					dataobj.text = format("0/0/%s", maxQuests)
				end
			else
				if core.numQuests > 0 then
					dataobj.text = format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedQuests/core.numQuests), core.numCompletedQuests, core.numQuests)
				else
					dataobj.text = "0/0"
				end
			end
		elseif prdb.showMaxQuest then
			dataobj.text = format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedQuests/maxQuests), core.numCompletedQuests, maxQuests)
		else
			dataobj.text = core.numCompletedQuests
		end
	elseif prdb.showNumQuests then
		if prdb.showMaxQuest then
			dataobj.text = format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numQuests/maxQuests), core.numQuests, maxQuests)
		else
			dataobj.text = core.numQuests
		end	
	elseif prdb.showMaxQuest then
		dataobj.text = maxQuests
	else
		dataobj.text = ""
	end
end

----------------------------
--- Custom Cell Provider ---
----------------------------
local cProvider, cPrototype, basecCellPrototype = LibQTip:CreateCellProvider(LibQTip.LabelProvider)
function cPrototype:InitializeCell()
	self.fontString = self:CreateFontString()
	self.fontString:SetFontObject(GameTooltipText)
end

function cPrototype:SetupCell(tooltip, value, justification, font, bType, bValue, maxWidth, indent, anchor)
	self:EnableMouse(true)
	self:SetScript("OnMouseDown", function(frame, button)
		if bType == "header" then
			if chdb.collapseAll == 1 then
				chdb.collapseAll = 0
				for zone in pairs(chdb.zoneDisplay) do
					chdb.zoneDisplay[zone] = 0
				end
			else
				chdb.collapseAll = 1
				for zone in pairs(chdb.zoneDisplay) do
					chdb.zoneDisplay[zone] = 1
				end
			end
			tip:SetScript("OnLeave", nil)
			LibQTip:Release(tip)
			tip = nil
			module:updateTooltip()
		elseif bType == "zone" then
			if chdb.zoneDisplay[bValue] == 1 then
				chdb.zoneDisplay[bValue] = 0
				chdb.collapseAll = 0
			else
				chdb.zoneDisplay[bValue] = 1
				chdb.collapseAll = 1
			end
			tip:SetScript("OnLeave", nil)
			LibQTip:Release(tip)
			tip = nil
			module:updateTooltip()
		elseif bType == "quest" then
			if button == "RightButton" then
				if ChatFrameEditBox:IsVisible() then
					if IsShiftKeyDown() then
						if core.quests[bValue].party[core.playerName] and core.quests[bValue].id and prdb.chatlink then
							ChatEdit_InsertLink(GetQuestLink(core.quests[bValue].id))
						end
					elseif IsControlKeyDown() then
						if core.quests[bValue].numObjectives > 0 and core.quests[bValue].party[core.playerName] and prdb.questsobjectives then
							for objective in pairs(core.questsObjectives[bValue]) do
								ChatFrameEditBox:Insert(format("{%s %s/%s} ", core.questsObjectives[bValue][objective].desc, core.questsObjectives[bValue][objective].party[core.playerName].got, core.questsObjectives[bValue][objective].need))
							end
						end
					end
				elseif core.quests[bValue].party[core.playerName] and core.quests[bValue].id and prdb.clicktooltip then
					ItemRefTooltip:ClearLines()
					ShowUIPanel(ItemRefTooltip)
					if not ItemRefTooltip:IsVisible() then
						ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
					end
					local link = GetQuestLink(core.quests[bValue].id)
					if link then
						ItemRefTooltip:SetHyperlink(link)
					end
				end
			elseif button == "LeftButton" then
				if IsControlKeyDown() then
					if core.numPartyMembers > 0 and core.quests[bValue].party[core.playerName] and core.quests[bValue].id and prdb.sharequest then
						local wasSelected = GetQuestLogSelection()
						SelectQuestLogEntry(core.quests[bValue].id)
						if GetQuestLogPushable() then
							QuestLogPushQuest()
						end
						SelectQuestLogEntry(wasSelected)
					end
				elseif IsShiftKeyDown() then
					if prdb.trackquest then
						core.cbh:Fire("toggleTrackedQuest", bValue)
					end
				elseif IsAltKeyDown() then
					if core.quests[bValue].party[core.playerName] and core.quests[bValue].id and prdb.abandonquest then
						local wasSelected = GetQuestLogSelection()
						SelectQuestLogEntry(core.quests[bValue].id)
						SetAbandonQuest()
						local items = GetAbandonQuestItems()
						if items then
							StaticPopup_Hide("ABANDON_QUEST")
							StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
						else
							StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
							StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
						end
						SelectQuestLogEntry(wasSelected)
					end
				else
					core:displayQuestInformations(bValue)
				end
			end
		end
	end)
	self:SetScript("OnEnter", function()
		if prdb.overtooltip and bType == "quest" and core.quests[bValue].id then
			local link = GetQuestLink(core.quests[bValue].id)
			if link and tooltip and anchor then
				GameTooltip:SetOwner(tooltip, anchor)
				GameTooltip:SetHyperlink(link)
			end
		end
	end)
	self:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)

	indent = indent or 0
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font, arg)
	local fs = self.fontString
	fs:SetPoint("TOPLEFT", self, "TOPLEFT", indent, 0)
	fs:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0)
	width = width + indent
	if width > maxWidth then
		fs:SetWidth(maxWidth)
		height = fs:GetHeight()
		width = maxWidth
	end
	fs:Show()
	return width, height
end

function cPrototype:ReleaseCell() 
	self:EnableMouse(false) 
	self:ClearAllPoints()
	self:SetScript("OnEnter", nil) 
	self:SetScript("OnLeave", nil) 
	self:SetScript("OnMouseDown", nil) 
end

function module:updateTooltip()
	local plus =  format("|TInterface\\Buttons\\UI-PlusButton-Up:%d|t", prdb.collapseButtonSize)
	local minus = format("|TInterface\\Buttons\\UI-MinusButton-Up:%d|t", prdb.collapseButtonSize)
	tip = LibQTip:Acquire("tomQuest2ToolTip", 2, "LEFT", "RIGHT")
	tooltipIsShown = true
	tip:Clear()
	tip:EnableMouse(true)
	tip:SetScript("OnLeave", hideTooltip)
	tip:SetScale(prdb.scale)
	local headerFont = CreateFont("headerFont")
	headerFont:SetFont(GameTooltipText:GetFont(), prdb.headerFontSize)
	tip:SetHeaderFont(headerFont)
	local smallFont = CreateFont("smallFont")
	smallFont:SetFont(GameTooltipText:GetFont(), prdb.normalFontSize)
	tip:SetFont(smallFont)
	local tempLine
	tempLine = tip:AddLine()
	tip:SetCell(tempLine, 1, format("%s |cffffffff%s|r", chdb.collapseAll == 0 and plus or minus,"tomQuest2"), headerFont, "CENTER", 2, cProvider, "header", nil, prdb.maxWidth)
	tip:AddLine(" ")
	for zone in pairs(core.questsZones) do
		chdb.zoneDisplay[zone] = chdb.zoneDisplay[zone] or 1
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, format("%s |cff%s%s|r |cff%s%s|r", chdb.zoneDisplay[zone] == 0 and plus or minus, prdb.colorizeHeaderName and core.questsZones[zone].color or "ffffff", zone, prdb.colorizeHeaderLevel and core.questsZones[zone].color or "ffffff", core.questsZones[zone].level ), headerFont, "LEFT", 2, cProvider, "zone", zone, prdb.maxWidth, 0)
		if chdb.zoneDisplay[zone] == 1 then
			for uid in pairs(core.questsByZone[zone].quests) do
				tempLine = tip:AddLine()
				tip:SetCell(tempLine, 1, format("|cff%s%s|r|cff%s%s|r %s", prdb.colorizeQuestLevel and core.quests[uid].color or "ffffff", core.quests[uid].tag, prdb.colorizeQuestName and core.quests[uid].color or "ffffff", core.quests[uid].title, core.quests[uid].playersOnQuest > 1 and format("<%d>", core.quests[uid].playersOnQuest) or ""), headerFont, "LEFT", 1, cProvider, "quest", uid, prdb.maxWidth, prdb.questscellpadding, tipAnchor)
				tip:SetCell(tempLine, 2, core.quests[uid].pqStatus, headerFont, "RIGHT", 1, cProvider, "quest", uid, prdb.maxWidth, 0, tipAnchor)
				if core.questsObjectives[uid] then
					for objective in pairs(core.questsObjectives[uid]) do
						if prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and not core.questsObjectives[uid][objective].partyComplete) or (prdb.hideCompletedObjective == 1 and not core.questsObjectives[uid][objective].playerComplete) then
							tempLine = tip:AddLine()
							tip:SetCell(tempLine, 1, format("|cff%s%s|r", prdb.colorizeQuestObjective and core.questsObjectives[uid][objective].qoColor or "ffffff", core.questsObjectives[uid][objective].desc), smallFont, "LEFT", 1, cProvider, "quest", uid, prdb.maxWidth, prdb.objectivescellpadding, tipAnchor)
							tip:SetCell(tempLine, 2, core.questsObjectives[uid][objective].qoStatus, smallFont, "RIGHT", 1, cProvider, "quest", uid, prdb.maxWidth, 0, tipAnchor)
						end
					end
				else
					tempLine = tip:AddLine()
					tip:SetCell(tempLine, 1, core.quests[uid].objective, smallFont, "LEFT", 2, cProvider,"quest", uid, prdb.maxWidth, prdb.objectivescellpadding, tipAnchor)
				end
			end
		end
	end
	tip:AddLine(" ")
	if prdb.questlog then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, LQTo["|cffeda55fLeft click|r |cff19ff19Open quests log.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxWidth)
	end
	if prdb.trackquest then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, LQTo["|cffeda55fShift + Left click|r |cff19ff19Add quest to tracker.|r"] , smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxWidth)
	end
	if prdb.sharequest then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, LQTo["|cffeda55fCtrl + Left click|r |cff19ff19Share quest.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxWidth)
	end
	if prdb.abandonquest then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, LQTo["|cffeda55fAlt+ Left click|r |cff19ff19Abandon quest.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxWidth)
	end
	if prdb.clicktooltip then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, LQTo["|cffeda55fRight click|r |cff19ff19Show quest tooltip.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxWidth)
	end
	if prdb.chatlink then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, LQTo["|cffeda55fShift + Right click|r |cff19ff19Send quest link to chat.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxWidth)
	end
	if prdb.questsobjectives then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, LQTo["|cffeda55fCtrl + Right click|r |cff19ff19Send quest progress to chat.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxWidth)
	end
	tip:Show()
end

function dataobj.OnEnter(self)
	tipAnchor = self:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	module:updateTooltip()
	tip:SmartAnchorTo(self)
end

function dataobj.OnLeave(self)
	hideTooltip()
end

function dataobj.OnClick(self,button)
	if button == "RightButton" then
		InterfaceOptionsFrame_OpenToCategory(core.mainOptionPanel)
	else
		ToggleFrame(QuestLogFrame)
	end
end